class UnexpectedStatusCode(Exception):
    """
    Raised when an api returns an unprocessable or unexpected status code
    """


class UnexpectedResourceResponse(Exception):
    """
    Raised when an api returns an unprocessable or unexpected response content
    """


from sieve.sdk import config


class BaseClientSession:
    connection_kwargs = {
        'limit': config.MAX_CONCURRENT_REQUESTS,
        'limit_per_host': config.MAX_CONCURRENT_REQUESTS_PER_HOST
    }



import asyncio
from typing import Dict

from aiohttp import ClientSession, TCPConnector
from sieve.sdk import config


class SemaphoredClientSession(ClientSession):
    default_connector_kwargs = {
        'limit': config.MAX_CONCURRENT_REQUESTS,
        'limit_per_host': config.MAX_CONCURRENT_REQUESTS_PER_HOST
    }

    def __init__(self, *,
                 connector=None,
                 connector_kwargs: Dict=None,
                 **kwargs):
        """
        A subclass of aiohttp's ClientSession which provides the possibility of
        limiting concurrent requests to some urls, based on a definition of an
        url prefix and a value to be limit concurrent requests.

        :param semaphores: An iterable of tuples (URL prefix, Semaphore value)
        to be be used by the session.
        :type semaphores: Iterable[Tuple[str, int]]
        :param connector_kwargs: A dict of keyword arguments to be passed to
        the ClientSession TCPConnector. By default, aiohttp creates a
        connector that limits concurrent connections to `100`. That limitation
        can also be overwritten by setting a new integer value to the
        environment variable `SIEVE_SDK_MAX_CONCURRENT_REQUESTS`.
        :type connector_kwargs: dict
        """
        semaphores = kwargs.pop('semaphores', [])

        if connector_kwargs is None:
            connector_kwargs = self.default_connector_kwargs
        if connector is None:
            connector = TCPConnector(**connector_kwargs)

        super().__init__(connector=connector, **kwargs)

        semaphore_value = config.MAX_CONCURRENT_REQUESTS
        self._default_semaphore = asyncio.Semaphore(semaphore_value)

        self._semaphores = tuple(
            (url_prefix, asyncio.Semaphore(value))
            for url_prefix, value in semaphores
        )

    async def _request(self, method, url, *args, **kwargs):
        semaphore = self.get_semaphore(url)
        async with semaphore:
            return await super()._request(method, url, *args, **kwargs)

    def get_semaphore(self, url: str) -> asyncio.Semaphore:
        for url_prefix, semaphore in self._semaphores:
            if url.startswith(url_prefix):
                return semaphore
        return self._default_semaphore



import unittest

from sieve.sdk.collections.document import NullableList, Document


class NullableListTests(unittest.TestCase):
    def test_deleting_an_item_makes_the_index_position_value_to_be_None(self):
        list_ = NullableList(('Xablau', 666, 'Xena'))
        del list_[1]
        assert list_ == ['Xablau', None, 'Xena']

    def test_it_deletes_items_by_index_string_representation(self):
        list_ = NullableList(('Xablau', 666, 'Xena'))
        del list_['1']
        assert list_ == ['Xablau', None, 'Xena']

    def test_its_possible_to_access_indexes_by_string_representation(self):
        list_ = NullableList(('Xablau', 666, 'Xena'))

        assert list_[0] == 'Xablau'
        assert list_['0'] == 'Xablau'


class DocumentTests(unittest.TestCase):
    def test_it_creates_elemets_if_path_doesnt_exists(self):
        doc = Document()
        doc['dogs.male'] = 'Xablau'

        assert doc == {'dogs': {'male': 'Xablau'}}

    def test_list_items_are_typecasted_to_NullableList_instances(self):
        doc = Document()
        doc['dogs'] = ['Xablau', 'Xena']

        assert isinstance(doc['dogs'], NullableList)
        assert doc == {'dogs': ['Xablau', 'Xena']}



import logging
from os import getenv

from simple_json_logger import JsonLogger


DATETIME_FORMAT = getenv("SIEVE_SDK_DATETIME_FORMAT", "%Y-%m-%d %H:%M:%S")
VERIFY_SSL = getenv('SIEVE_SDK_VERIFY_SSL', '1') == '1'
LOG_LEVEL = int(getenv('SIEVE_LOG_LEVEL', logging.DEBUG))
MAX_CONCURRENT_REQUESTS = int(getenv('SIEVE_SDK_MAX_CONCURRENT_REQUESTS', 200))
MAX_CONCURRENT_REQUESTS_PER_HOST = int(getenv('SIEVE_SDK_MAX_CONCURRENT_REQUESTS_PER_HOST', 0))
PROXY_ADDRESS = getenv('SIEVE_PROXY_ADDRESS', 'socks5://127.0.0.1:8080')
PROXY_SHOULD_REMOTE_RESOLVE = getenv('SIEVE_SDK_PROXY_SHOULD_REMOTE_RESOLVE', '1') == '1'

logger = JsonLogger(level=LOG_LEVEL)



import abc
import asyncio
from http import HTTPStatus
from typing import Optional, Dict, List, Union, Iterable

from aiohttp import ClientSession, web

from sieve.sdk.config import VERIFY_SSL, logger
from sieve.sdk.exceptions import UnexpectedStatusCode


class ABCApiResource(metaclass=abc.ABCMeta):
    resource_url: str = NotImplementedError

    headers: Optional[Dict] = None

    def __init__(self, session: ClientSession, verify_ssl: bool=None):
        self.session = session

        if verify_ssl is None:
            verify_ssl = VERIFY_SSL
        self.verify_ssl = verify_ssl

    def url_for_id(self, id_) -> str:
        return f"{self.resource_url}/{id_}"

    async def search(self, params: Dict) -> Optional[Union[Dict, List[Dict]]]:
        async with self.session.get(url=self.resource_url,
                                    params=params,
                                    headers=self.headers,
                                    verify_ssl=self.verify_ssl) as response:
            if response.status == HTTPStatus.OK:
                return await response.json()
            elif response.status == HTTPStatus.NOT_FOUND:
                return None

            response_body = await response.text()

            logger.error({
                'info': 'Unable to search resource',
                'response_body': response_body,
                'response_status': response.status,
                'request_url': response.url.human_repr()
            })
            raise UnexpectedStatusCode({
                "status": response.status,
                "body": response_body
            })

    async def search_many(self, params: Iterable[Dict], return_exceptions=False):
        """
        Accessory method for concurrently doing multiple searchs with gather
        """
        tasks = (self.search(p) for p in params)
        return await asyncio.gather(*tasks,
                                    return_exceptions=return_exceptions)

    async def get(self, id_) -> Optional[Dict]:
        url = f'{self.resource_url}/{id_}'
        async with self.session.get(url=url,
                                    headers=self.headers,
                                    verify_ssl=self.verify_ssl) as response:
            if response.status == HTTPStatus.OK:
                return await response.json()
            elif response.status == HTTPStatus.NOT_FOUND:
                return None

            response_body = await response.text()
            logger.error({
                'info': 'Unable to get resource',
                'response_body': response_body,
                'response_status': response.status,
                'request_url': response.url.human_repr()
            })
            raise UnexpectedStatusCode({
                "status": response.status,
                "body": response_body
            })

    async def get_many(self,
                       ids: List[int],
                       return_exceptions=False) -> List[Dict]:
        """
        Accessory method for concurrently doing multiple ABCApiResource.get
        using gather
        """
        tasks = (self.get(id_) for id_ in ids)
        results = await asyncio.gather(*tasks,
                                       return_exceptions=return_exceptions)
        return results

    def _create_response_is_valid(self, response: web.Response) -> bool:
        return response.status == HTTPStatus.CREATED

    async def create(self, data: Dict) -> Dict:
        """
        Performs a POST request on the remote resource

        :param data: A dict of the paramenters of the resource to be created
        :return:
        """
        async with self.session.post(url=self.resource_url,
                                     headers=self.headers,
                                     json=data,
                                     verify_ssl=self.verify_ssl) as response:
            if self._create_response_is_valid(response):
                return await response.json()

            response_body = await response.text()
            logger.error({
                "body": data,
                "url": response.url.human_repr(),
                "response_status": response.status,
                "response": response_body,
                "info": "Unable to create resource"
            })
            raise UnexpectedStatusCode({
                "status": response.status,
                "body": response_body
            })

    def _update_response_is_valid(self, response: web.Response) -> bool:
        return response.status == HTTPStatus.OK

    async def update(self, id_: int, params: Dict) -> bool:
        async with self.session.put(url=f"{self.resource_url}/{id_}",
                                    headers=self.headers,
                                    json=params,
                                    verify_ssl=self.verify_ssl) as response:
            if self._update_response_is_valid(response):
                return True

            response_body = await response.text()
            logger.error({
                "params": params,
                "url": response.url.human_repr(),
                "response_status": response.status,
                "response": response_body,
                "info": "Unable to update resource data"
            })
            raise UnexpectedStatusCode({
                "status": response.status,
                "body": response_body
            })

    def _delete_response_is_valid(self, response: web.Response) -> bool:
        return response.status == HTTPStatus.OK

    async def delete(self, id_: int) -> bool:
        async with self.session.delete(url=f"{self.resource_url}/{id_}",
                                       headers=self.headers,
                                       verify_ssl=self.verify_ssl) as response:
            if self._delete_response_is_valid(response):
                return True

            response_body = await response.text()
            logger.error({
                "url": response.url.human_repr(),
                "response_status": response.status,
                "response": response_body,
                "info": "Unable delete resource"
            })
            raise UnexpectedStatusCode({
                "status": response.status,
                "body": response_body
            })



import json
import datetime
from functools import partial

from sieve.sdk import config
from pathdict.collection import StringIndexableList, PathDict


def default_handler(value):
    if isinstance(value, datetime.datetime):
        return value.strftime(config.DATETIME_FORMAT)
    elif isinstance(value, PathDict):
        return dict(value)
    elif isinstance(value, StringIndexableList):
        return list(value)
    raise TypeError("Unexpected type")


dumps = partial(json.dumps, default=default_handler)



from pathdict.collection import StringIndexableList, PathDict


class NullableList(StringIndexableList):
    def __delitem__(self, key):
        if isinstance(key, str):
            key = int(key)
        self[key] = None


class Document(PathDict):
    """
    A PathDict subclass that automatically creates non existing keys in the
    path, and creates list elements as NullableList instances.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs,
                         create_if_not_exists=True,
                         list_class=NullableList)



import warnings
from functools import partial
from typing import Dict

from aiohttp import ClientSession, ClientRequest
from aiosocksy.connector import ProxyConnector, ProxyClientRequest
from yarl import URL

from sieve.sdk import config


class ProxiedRequest(ProxyClientRequest):
    def __init__(self, method, url, *,
                 proxy=None,
                 proxy_address=None,
                 **kwargs):
        if proxy is None:
            proxy = proxy_address

        super(ProxiedRequest, self).__init__(method,
                                             url,
                                             proxy=URL(proxy),
                                             **kwargs)


class Socks5ProxiedClientSession(ClientSession):
    default_connector_kwargs = {
        'limit': config.MAX_CONCURRENT_REQUESTS,
        'limit_per_host': config.MAX_CONCURRENT_REQUESTS_PER_HOST
    }

    """
    An aiohttp.ClientSession mixin that provides transparent socks5 proxy
    integration
    """

    def __init__(self, *,
                 connector=None,
                 connector_kwargs: Dict=None,
                 request_class: ClientRequest=None,
                 proxy_address: str=None,
                 remote_resolve: bool=None,
                 **kwargs):
        """
        :type proxy_address: str
        :param proxy_address: Address of the proxy to be used.
        Ex.:
            Socks5ProxiedClientSession(proxy_address='socks5://127.0.0.1:8080')
            Socks5ProxiedClientSession(proxy_address='http://127.0.0.1:1080')

        :param remote_resolve: Whether or not to do local domain name resolving
        :type remote_resolve: bool
        """
        if proxy_address is None:
            warnings.warn("Creating a session without explicitly "
                          "calling it with proxy_address. Using default.")
            proxy_address = config.PROXY_ADDRESS
        self.proxy_address = proxy_address

        if remote_resolve is None:
            remote_resolve = config.PROXY_SHOULD_REMOTE_RESOLVE
        self.remote_resolve = remote_resolve

        if connector_kwargs is None:
            connector_kwargs = self.default_connector_kwargs
        if connector is None:
            connector = ProxyConnector(remote_resolve=remote_resolve,
                                       **connector_kwargs)

        if request_class is None:
            request_class = partial(ProxiedRequest,
                                    proxy_address=self.proxy_address)

        super().__init__(connector=connector,
                         request_class=request_class,
                         **kwargs)



import os
from setuptools import setup, find_packages
from sieve.sdk import __VERSION__


long_description = open(os.path.join(os.path.dirname(__file__), 'README.md'),
                        encoding='utf-8').read()

setup(
    name='sieve_sdk',
    version=__VERSION__,
    description='Sieve SDK',
    long_description=long_description,
    author='Diogo Magalhaes Martins',
    author_email='diogo.martins@sieve.com.br',
    url='https://bitbucket.org/sievetech/sieve-python-sdk',
    packages=find_packages(exclude=['tests', 's3_upload', 'deploy.sh']),
    install_requires=[
        "aiodns==1.1.1",
        "aiohttp>=3.1.2",
        "aiosocksy==0.1",
        "cchardet==2.1.1",
        "pathdict>=0.0.9",
        "simple_json_logger>=0.2.2"
    ],
    include_package_data=True,
)
